#
# _printer.py
#
# A simple printer class
#
# by Nick Patavalis (npat@efault.net)
#
# This is part of the xmlmp program which is free software; you can
# redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# The xmlmp program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the xmlmp program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA
#

import sys

class Printer:

    """ A Simple printer.

    Prints to file-objects. Removes unnecessary blanks from the input
    stream, fills lines, optionally wraps long lines, optionally
    indents lines, and can also accept preformated chuncks of text
    that are printed verbatim.

    """
    
    def __init__ (self, wrap = 1, linelen = 70, out = sys.stdout) :

        """ Create a printer-instance

        Takes three optional named (keyword) arguments. 'wrap' is a
        flag indicating if lines must be wrapped or not. 'linelen' is
        used only if 'wrap' is set, and defines the line length. 'out'
        is the file object to emmit (print) the text to.

        """

        self._out = out

        # _pos is the current offset from the beginning of the
        # line. used for wrapping long lines
        
        self._pos = 0

        # if _cword is set, then the last word emitted is not a full
        # word, but has to be completed by text from the next "pr" or
        # "pr_pre" invocation. For example after pr("this is a te")
        # _cword will be 1 [expecting the pr("st.")], while after
        # pr("This is a\n") _cword will be 0.
        
        self._cword = 0

        # _nl defines the string emitted by the printer as a
        # line-termination indication.
        
        self._nl = '\n'

        # _lsep is the set of characters that constitute line
        # terminators in the *input* stream.
        
        self._lsep = '\r\n'

        # _wsep is the set of characters that constitute word
        # separators in the *input* stream. In the output stream words
        # are always separated by a singe space or a _nl.
        
        self._wsep = '\r\n\t '

        # current indentation level (expressed in number of characters
        # from the beginning of the line)
        
        self._indent = 0
        
        self._wrap = wrap
        self._linelen = linelen


    def pr_pre(self, s, do_indent = 1):

        """ Print a preformated chunk of text.

        's' is the text to print, and 'do_indent' indicates whether
        lines should be indented according to the current indentation
        level, or not. An interesting quirck, it that if the
        preformated text starts with a line-termination character,
        then this character is treated as an 'open-line' request. that
        is, if the current position is already at the begining of a
        line, then the initial line-terminating character will be
        ignored. For example:

           pr_pre('foo bar')
           pr_pre('\\nfoobar\\n')
           pr_pre('\\nfoobarbaz\\n')

        will produce:

          foo bar
          foobar
          foobaz

        Also, if line-wrapping is on, then the first line of the
        preformated text will be subjected to the normal wrapping
        rules, as if the were a *single* word. For example, assuming
        that the current line is like this:
        
                                           wrap here
                                           |
                                           V
          ----------------------------------
          one two three four five six

        Then, pr_pre('foo bar\\n' + 'foobar' * 6), will result to:

                                           wrap here
                                           |
                                           V
          ----------------------------------
          one two three four five six
          foo bar
          foobar foobar foobar foobar foobar foobar

        While, in the same situation,
        pr_pre('foo\\n' + 'foobar' * 6)
        will result to:
        
                                           wrap here
                                           |
                                           V
          ----------------------------------
          one two three four five six foo
          foobar foobar foobar foobar foobar foobar
          
        """
        
        lines = s.splitlines()
        nlines = len(lines)
        if not nlines :
            return
        # is the last line terminated?
        hasnl = not self._lsep.find(s[-1]) < 0
        
        # handle the first line. a single line is considered
        # to be the first line, not the last one.
        l = lines[0]
        if l == '' :
            if ( self._pos != 0 ):
                self._out.write(self._nl)
                self._pos = 0;
            if nlines == 1 :
                self._cword = 0                
                return
        else :
            if not self._cword :
                if ( self._wrap
                     and self._pos
                     and self._pos + len(l) + 1 > self._linelen ) :
                    self._out.write(self._nl)
                    self._pos = 0
                elif self._pos != 0 :
                    self._out.write(' ')
                    self._pos += 1
            if self._pos == 0 :
                self._out.write(' ' * self._indent)
                self._pos += self._indent
            self._out.write(l)
            self._pos += len(l)
            if nlines == 1 :
                if hasnl :
                    self._out.write(self._nl)
                    self._pos = 0
                    self._cword = 0
                else :
                    self._cword = 1
                return

        # handle the lines in the middle
        for l in lines[1:-1]:
            if ( do_indent ):
                self._out.write(' ' * self._indent + l + self._nl)
            else :
                self._out.write(l + self._nl)
            self._pos = 0

        # handle the last line
        l = lines[-1]
        if ( do_indent ):
            self._out.write(' ' * self._indent + l)
            self._pos = self._indent + len(l)
        else :
            self._out.write(l)
            self._pos = len(l)
        if hasnl :
            self._out.write(self._nl)
            self._pos = 0
            self._cword = 0
        else :
            self._cword = 1

    def pr(self, s):

        """ Print a chunk of text

        's' is the text to print. Unnecessary blanks are removed. In
        the output stream, every word is separated from the next by a
        single space character. If the printer was initialized with
        'wrap=1' then lines are wrapped when they become longer than
        the value specified by the 'linesize' initialization
        parameter. The input stream is treated as a list of words
        separated by [\\r\\n\\t ]. The end and the beginning of the
        chunks are *not* considered word-separators. For example:

          pr('one two three')
          pr('four five six')

        will produce:

          one two threefour five six

        """
        
        words = s.split()
        if not self._wsep.find(s[0]) < 0 :
            self._cword = 0
        hassep = not self._wsep.find(s[-1]) < 0
        for w in words:
            if not self._cword :
                if ( self._wrap
                     and self._pos
                     and self._pos + len(w) + 1 > self._linelen ) :
                    self._out.write(self._nl)
                    self._pos = 0
                elif self._pos :
                    self._out.write(' ')
                    self._pos += 1
            if self._pos == 0 :
                self._out.write(' ' * self._indent)
                self._pos += self._indent
            self._out.write(w)
            self._pos += len(w)
            self._cword = 0
        if not hassep :
            self._cword = 1

    def indent (self, offset) :

        """ Change the current indentation level

        'offset' is the amount to change the indendation level by,
        expressed in characters. If 'offset' is positive, the
        indentation level increases, if it is negative it
        decreases. The new indentation level will apply when a new
        line is opened.
        
        """
        
        self._indent += offset

